
# Main Sections:
#     Changing the tree topology: adding nodes, reparenting nodes
#     Tree traversal
#     Making grid versions of the tree for display and growth
#     Utility functions: Potential function, etc
#     Filling in costs of veins: weights, bends, accumulation
#     Vein optimisation
#     Adding cells at border
#     Display, as image and text
#     Examples
#     Top level 


################################################################################
# Changing the tree topology: adding nodes, reparenting nodes

make.root <- function() {
	root = list(N=1, pN=-1, cNs=integer(0), x=0, y=0, w=1, m=1, f=1, F=1)
	list(root)
}


make.node <- function() {
	list(N=-1, pN=-1, cNs=integer(0), x=NA, y=NA, w=NA, m=NA, f=NA, F=NA)
}


add.child <- function(tree, pN, x, y) {
	node = make.node()
	node$N = length(tree) + 1
	node$pN = pN
	node$x = x
	node$y = y
	tree[[pN]]$cNs = c(tree[[pN]]$cNs, node$N)
	new.tree = c(tree , list(node))
	new.tree
}



# can reparent to existing parent (does nothing)
reparent.node <- function(tree, n, qN) {
	# remove n from current parent's children
	pN = tree[[n]]$pN
	pcNs = tree[[pN]]$cNs
	tree[[pN]]$cNs = pcNs[pcNs != n]
	# make qN the new parent
	tree[[n]]$pN = qN
	tree[[qN]]$cNs = c(tree[[qN]]$cNs, n)
	tree
}


################################################################################
# Tree traversal

collect.subtree.postorder <- function(tree, N, Ns) {
	node = tree[[N]]
	nc = length(node$cNs)
	if (nc > 0) {
		for (ch in 1:nc) {
			Ns = collect.subtree.postorder(tree, node$cNs[ch], Ns)
		}
	}
	Ns = c(Ns, node$N)
	Ns
}


collect.subtree.preorder <- function(tree, N, Ns) {
	node = tree[[N]]
	Ns = c(Ns, node$N)
	nc = length(node$cNs)
	if (nc > 0) {
		for (ch in 1:nc) {
			Ns = collect.subtree.preorder(tree, node$cNs[ch], Ns)
		}
	}
	Ns
}

collect.postorder <- function(tree) {
	Ns = NULL
	Ns = collect.subtree.postorder(tree, 1, Ns)
	Ns
}  


collect.preorder <- function(tree) {
	Ns = NULL
	Ns = collect.subtree.preorder(tree, 1, Ns)
	Ns
}  


################################################################################
# making grid versions of the tree for both display, and finding
# cells next to border to (possibly) add.

# Note that grid is (row, column) so (y,x) and image has rows increasing
# upwards, unlike matrix
# root is at (0,0) in tree, (1-miny, 1-minx) in grid


tree.bbox <- function(tree) {
	minx = Inf
	maxx = -Inf
	miny = Inf
	maxy = -Inf
	nnodes = length(tree)
	for (n in 1:nnodes) {
		minx = min(minx, tree[[n]]$x)
		maxx = max(maxx, tree[[n]]$x)
		miny = min(miny, tree[[n]]$y)
		maxy = max(maxy, tree[[n]]$y)
	}
	list(minx=minx, maxx=maxx, miny=miny, maxy=maxy, w = 1 + maxx - minx, h = 1 + maxy - miny) 
}

x.y.to.N <- function(grid, bbox, x, y) {
	xx = 1 + x - bbox$minx
	yy = 1 + y - bbox$miny
	n = -1
	if (xx >= 1 & xx <= bbox$w & yy >= 1 & yy <= bbox$h )
	{
		n = grid[yy,xx]
	}
	n
}



tree.as.grid.nodenums <- function(tree, bbox) {
	grid = matrix(-1, nrow=bbox$h, ncol=bbox$w)
	nnodes = length(tree)
	for (n in 1:nnodes) {
		grid[ 1 + tree[[n]]$y - bbox$miny, 1 + tree[[n]]$x - bbox$minx] = tree[[n]]$N
	}
	grid
}


tree.as.grid.Fs <- function(tree, bbox) {
	grid = matrix(0, nrow=bbox$h, ncol=bbox$w)
	nnodes = length(tree)
	maxF = tree[[1]]$F
	for (n in 1:nnodes) {
		grid[ 1 + tree[[n]]$y - bbox$miny, 1 + tree[[n]]$x - bbox$minx] = tree[[n]]$F / maxF
	}
	grid
}



################################################################################
#     Utility functions: Potential function, etc




dist.between.nodes <- function(tree, node, node2) {
	x = node$x
	y = node$y
	px = node2$x
	py = node2$y
	dx = x - px
	dy = y - py
	dist = sqrt(dx*dx+dy*dy)	
	dist
}


dist.to.node <- function(tree, node) {
	pN = node$pN
	dist = 1 # for root
	if (pN > 0) {
		dist = dist.between.nodes(tree, node, tree[[pN]])
	}
	dist	
}



# Potential function
# list(N=-1, pN=-1, cNs=integer(0), x=NA, y=NA, w=NA, m=NA, f=NA, F=NA)
potential <- function(tree, params, n, dw) {
	pp = 0
	nn = n
	alpha = params["alpha"]
	beta = params["beta"]
	repeat {
		node = tree[[nn]]
		pnn = tree[[nn]]$pN
		if (pnn <= 0) {
			# got to root
			break
		}
		d = dist.to.node(tree, node)
		w = node$w
		m = node$m
		pp = pp + m^beta * ((w+dw)^alpha - w^alpha) * d
		nn = pnn
	}
	pp
}


is.descendant.of <- function(tree, q, n) {
	descendant = FALSE
	nn = q
	repeat {
		if (nn == n) {
			descendant = TRUE
			break
		}
		nn = tree[[nn]]$pN
		if (nn <= 0) {
			# got to root
			break
		}
	}
	descendant
}


################################################################################
# Filling in costs of veins: weights, bends, accumulation

fillin.weights <- function(tree) {
	Ns = collect.postorder(tree)
	for (N in Ns) {
		node = tree[[N]]
		w = 1
		nc = length(node$cNs)
		if (nc > 0) {
			for (ch in 1:nc) {
				cN = node$cNs[ch]
				w = w + tree[[cN]]$w
			}
		}
		tree[[N]]$w = w
	}
	tree
}


bendcost <- function(pdx, pdy, dx, dy) {
	cos.theta = 0
	dotp = pdx*dx + pdy*dy
	if (dotp > 0) {
		dist = sqrt((pdx^2 + pdy^2) * (dx^2 + dy^2))
		cos.theta = dotp / dist
	}
	1 / cos.theta
}



# c, p, g play the role of child, parent, grandparent but don't 
# have to have that relationship in the tree. c and p cannot be root.
# used for a child of a node being reparented (so child is regrandparented)
three.node.bendcost <- function(tree, c, p, g) {
	node = tree[[c]]
	x = node$x
	y = node$y
	px = tree[[p]]$x
	py = tree[[p]]$y
	ppx = tree[[g]]$x
	ppy = tree[[g]]$y
	pdx = px - ppx
	pdy = py - ppy
	dx = x - px
	dy = y - py
	b = bendcost(pdx, pdy, dx, dy)		
	b
}


# n, p play the role of node, parent, but don't 
# have to have that relationship in the tree. It returns 1 if n or
# p is the root.
node.bendcost <- function(tree, n, p) {
	node = tree[[n]]
	b = 1 # for root
	if (p > 0) {
		x = node$x
		y = node$y
		px = tree[[p]]$x
		py = tree[[p]]$y
		# parent coords of root in case no grandparent
		ppx = Stalk.x
		ppy = Stalk.y
		pp = tree[[p]]$pN
		if (pp > 0) {
			ppx = tree[[pp]]$x
			ppy = tree[[pp]]$y
		}
		pdx = px - ppx
		pdy = py - ppy
		dx = x - px
		dy = y - py
		b = bendcost(pdx, pdy, dx, dy)		
	}
	b
}


fillin.bends <- function(tree) {
	Ns = collect.preorder(tree)
	for (n in Ns) {
		pN = tree[[n]]$pN
		m = 1
		if (pN > 0) {
			b = node.bendcost(tree, n, pN)
		  m = tree[[pN]]$m * b
		}
		tree[[n]]$m = m
	}
	tree
}




fillin.vein.costs <- function(tree, alpha, beta) {
	nnodes = length(tree)
	for (N in 2:nnodes) {
		node = tree[[N]]
		pN = node$pN
		stopifnot(pN > 0)
		dx = node$x - tree[[pN]]$x
		dy = node$y - tree[[pN]]$y
		tree[[N]]$f = node$m^beta * node$w^alpha * sqrt(dx^2 + dy^2)
	}
	Ns = collect.postorder(tree)
	for (N in Ns) {
		node = tree[[N]]
		F = tree[[N]]$f
		nc = length(node$cNs)
		if (nc > 0) {
			for (ch in 1:nc) {
				cN = node$cNs[ch]
				F = F + tree[[cN]]$F
			}
		}
		tree[[N]]$F = F
	}
	tree
}



fillin.all.costs <- function(tree, params) {
	tree = fillin.weights(tree)
	tree = fillin.bends(tree)
	tree = fillin.vein.costs(tree, params["alpha"], params["beta"])
	tree
}



################################################################################
# Vein optimisation

# if (x,y) was not in tree, qN will be < 0, not valid
# not valid if qN is descendant of n 
# not valid if bend (parent(qN) to qN to n) is too sharp
# not valid if bend (qN to n to a child) is too sharp
new.parent.is.valid <- function(tree, params, grid, bbox, n, qN) {
	valid = FALSE
  if (qN > 0) {
  	desc = is.descendant.of(tree, qN, n)
  	if (!desc) {
  		b.q = node.bendcost(tree, n, qN)
  		if (is.finite(b.q)) {
  			node = tree[[n]]
  			childrenok = TRUE
  			for (c in node$cNs) {
  				bc.q = three.node.bendcost(tree, c, n, qN)
  				if (!is.finite(bc.q)) {
  					childrenok = FALSE
  				}
  			}  				
        if (childrenok) {
        	valid = TRUE
        }
  		}
  	}
  }
	valid
}


delta.cost.paths <- function(tree, params, grid, bbox, n, qN) {
	node = tree[[n]]
	pN = node$pN
  dcost = potential(tree, params, qN, node$w)
  stopifnot(is.finite(dcost) | is.infinite(dcost))
  dcost = dcost - potential(tree, params, pN, node$w)
  stopifnot(is.finite(dcost) | is.infinite(dcost))	
  dcost
}


delta.cost.node <- function(tree, params, grid, bbox, n, qN) {
	node = tree[[n]]
	pN = node$pN
	beta = params["beta"]
	dcost = Inf # in case bends are too sharp
	b.p = node.bendcost(tree, n, pN)
	stopifnot(is.finite(b.p))
	b.q = node.bendcost(tree, n, qN)
	stopifnot(is.finite(b.q))
	d.p = dist.to.node(tree, node)
	d.q = dist.between.nodes(tree, node, tree[[qN]])
  dcost = ( (b.q/b.p)^beta * (d.q/d.p) - 1 ) * node$f
	stopifnot(is.finite(dcost))
	dcost
}



delta.cost.subtrees <- function(tree, params, grid, bbox, n, qN) {
	dcost = 0
	node = tree[[n]]
	pN = node$p
	beta = params["beta"]
	if (length(node$cNs) > 0) {
		for (c in node$cNs) {
			bn.p = node.bendcost(tree, n, pN)
			bn.q = node.bendcost(tree, n, qN)
			bc.p = three.node.bendcost(tree, c, n, pN)
			bc.q = three.node.bendcost(tree, c, n, qN)
			if (!is.finite(bc.q)) {
				cat("c, n, qN", c, n, qN, "\n")
			}
			stopifnot(is.finite(bc.q))
			dcost = dcost + (((bn.q*bc.q)/(bn.p*bc.p))^beta - 1) * tree[[c]]$F
		}		
	}
	stopifnot(is.finite(dcost))
	dcost
}



cost.reparent.to.x.y <- function(tree, params, grid, bbox, n, x, y) {
  dcost = Inf 
  qN = x.y.to.N(grid, bbox, x, y)
  if (new.parent.is.valid(tree, params, grid, bbox, n, qN)) {  
		dcost = delta.cost.paths(tree, params, grid, bbox, n, qN)
		dcost = dcost + delta.cost.node(tree, params, grid, bbox, n, qN)
		dcost = dcost + delta.cost.subtrees(tree, params, grid, bbox, n, qN)
	}
  stopifnot(is.finite(dcost) | is.infinite(dcost))
	dcost
}




tryto.improve.parent <- function(tree, params, grid, bbox, n) {
	node = tree[[n]]
	pN = node$pN
	x = tree[[n]]$x
	y = tree[[n]]$y
	pN = tree[[n]]$pN
	
	stopifnot(is.finite(pN))
	stopifnot(pN>0)
	
	stopifnot(is.finite(pN))
	stopifnot(pN > 0)
	px = tree[[pN]]$x
	py = tree[[pN]]$y
	minx = min(x, px) - 1
	maxx = max(x, px) + 1
	miny = min(y, py) - 1
	maxy = max(y, py) + 1
	best.xx = NA
	best.yy = NA
	best.cost = 0
	for (xx in minx:maxx) {
		for (yy in miny:maxy) {
			if ((abs(xx-x) <= 1 & abs(yy-y) <= 1)  |  (abs(xx-px) <= 1 & abs(yy-py) <= 1)) {
				cost = cost.reparent.to.x.y(tree, params, grid, bbox, n, xx, yy)
				stopifnot(is.finite(cost) | is.infinite(cost))
				stopifnot(is.finite(best.cost))
				if (cost < best.cost) {
					best.cost = cost
					best.xx = xx
					best.yy = yy
				}
			}
		}
	}
	best.new.tree = tree
	if (best.cost < 0) {
	  best.new.tree = reparent.node(tree, n, x.y.to.N(grid, bbox, best.xx, best.yy))
	}
	best.new.tree
}



tryto.improve.veins <- function(tree, params, grid, bbox) {
	nnodes = length(tree)
	for (n in 2:nnodes) {
		tree = tryto.improve.parent(tree, params, grid, bbox, n)
	}
	tree
}



optimise.veins <- function(tree, params) {
	bbox = tree.bbox(tree)
	grid = tree.as.grid.nodenums(tree, bbox)
	
	repeat {
		F1 = tree[[1]]$F
		tree = tryto.improve.veins(tree, params, grid, bbox)
		tree = fillin.all.costs(tree, params)
		F2 = tree[[1]]$F
		if (F2 >= F1) break
	}
	tree
}




################################################################################
# Adding cells at border

best.parent <- function(tree, grid, bbox, params, x, y) {
	pNs = integer(0)
	for (xx in (x-1):(x+1)) {
		for (yy in (y-1):(y+1)) {
			pN = x.y.to.N(grid, bbox, xx, yy)
			if (pN > 0) {
				# (xx,yy) is in tree
				pNs = c(pNs, pN)
			}
		}
	}
	bestp = -1
	bestcost = Inf
	if (length(pNs) > 0) {
		beta = params["beta"]
		besti = -1
		for (i in 1:length(pNs)) {
			pn = pNs[i]
			cost = potential(tree, params, pn, 1)
			px = tree[[pn]]$x
			py = tree[[pn]]$y
			ppn = tree[[pn]]$pN
			ppx = Stalk.x
			ppy = Stalk.y # in case pn is root
			if (ppn > 0) {
				ppx = tree[[ppn]]$x
				ppy = tree[[ppn]]$y
			}
			dx = x-px
			dy = y-py
			b = bendcost(px-ppx, py-ppy, dx, dy)
			m = tree[[pn]]$m
			d = sqrt(dx*dx+dy*dy)
			cost = cost + (m*b)^beta * d
			if (cost < bestcost) {
				bestcost = cost
				besti = i
			}
		}
		if (besti > 0) {
			bestp = pNs[besti]
		}
	}
	list(p=bestp, cost=bestcost, x=x, y=y)
}



list.cells.to.add <- function(tree, params) {
	bbox = tree.bbox(tree)
	grid = tree.as.grid.nodenums(tree, bbox)
	new.tree = tree
	minx = bbox$minx - 1
	maxx = bbox$maxx + 1
	miny = bbox$miny - 1  
	maxy = bbox$maxy + 1
	cells = list()
	for (x in minx:maxx) {
		for (y in miny:maxy) {
			m = x.y.to.N(grid, bbox, x, y)
			if (m < 0) {
				# (x,y) is not in tree
				best = best.parent(tree, grid, bbox, params, x, y)
				if ((best$p > 0)) {
					cells = c(cells, list(best))
				}
			}
		}
	}
	mincost = Inf
	if (length(cells) > 0) {
		for (i in 1:length(cells)) {
			if (cells[[i]]$cost < mincost) {
				mincost = cells[[i]]$cost
			}
		}		
		cheapestcells = list()
		for (i in 1:length(cells)) {
			if (cells[[i]]$cost < mincost + 0.00001) {
				cheapestcells = c(cheapestcells, list(cells[[i]]))
			}
		}
		cells = cheapestcells
	}
  cells
}



add.cells.via.list <- function(tree, params) {
	cells = list.cells.to.add(tree, params)
	bbox = tree.bbox(tree)
	new.tree = tree
	numadded = 0
	cost = Inf
	if (length(cells) > 0) {
		cost = cells[[1]]$cost
		for (i in 1:length(cells)) {
			if (cells[[i]]$cost < params["epsilon"]) {
				new.tree = add.child(new.tree, cells[[i]]$p,  cells[[i]]$x,  cells[[i]]$y)
				numadded = numadded + 1
			}
		}		
	}
	list(new.tree=new.tree, added=numadded, cost=cost)
}
	


################################################################################
# Display, as image and text

tree.as.image <- function(tree, h, w, rooth) {
	bbox = tree.bbox(tree)
	grid = tree.as.grid.Fs(tree, bbox)
	plotgrid = matrix(0, nrow=h, ncol=w)
	x = 1:w
	y = 1:h
	offx = (w-1)/2
	offy = rooth
	for (r in 1:bbox$h) {
		for (c in 1:bbox$w) {
			plotgrid[r+offy + bbox$miny, c+offx + bbox$minx] = grid[r,c]
		}
	}
	z = t(plotgrid)^.33
	nc = 100
	clrs = rev(hcl.colors(nc, palette = "viridis", alpha = NULL, rev = FALSE, fixup = TRUE))
	clrs = c("#f8f8f8", clrs)
	image(x, y, z, col=clrs, axes=FALSE, ann=FALSE, xlim=c(0,w)+0.5, ylim=c(0,h)+0.5)
	nnodes = length(tree)
	for (n in 2:nnodes) {
		node = tree[[n]]
		pnode = tree[[ node$pN ]]
		lines(x=1 + offx+c(pnode$x, node$x), y=1 + offy+c(pnode$y, node$y), lwd = 0.5*sqrt(node$w))
	}
}



tree.nodes.as.matrix <- function(tree) {
	nnodes = length(tree)
	mm = matrix(0, nrow=nnodes, ncol=8)
	colnames(mm) = c("N","pN","x","y","W","m","f","F")
	for (n in 1:nnodes) {
		node = tree[[n]]
		mm[n,] = c(node$N, node$pN, node$x, node$y, node$w, node$m, node$f, node$F)
	} 
	mm
}


print.nodes <- function(tree) {
	mm = tree.nodes.as.matrix(tree)
	print(mm)
}



save.image <- function(tree, params, h, w, rooth) {
	fname = paste(as.character(100*params), collapse='-')
	fname = paste0(fname, ".png")
	png(filename = fname, width = 20*w, height = 20*h, units = "px", pointsize = 12, bg = "white")
	par(mar=c(0,0,0,0))
	tree.as.image(tree, h, w, rooth)
	dev.off()
}



run.and.display <- function(params, h, w, root.y) {
	tree = make.root()
	repeat {
		growth = add.cells.via.list(tree, params)
		tree = growth$new.tree
		tree = fillin.all.costs(tree, params)
		
		cat("added", growth$added, "at cost", growth$cost, "to make", length(tree), "cells\n")
		if (growth$added == 0) {
			break
		}
		
		tree = optimise.veins(tree, params)
		tree.as.image(tree, h, w, root.y)
	}
	save.image(tree, params, h, w, root.y)
	tree	
}





###################################################################
# examples - alpha, beta, epsilon  and image height, width, root position

example.55.100.50 <- function() {
	params = c(alpha = .55, beta = 1.00, epsilon = 5)
	run.and.display(params, 40, 29, 1)
}

example.55.30.50 <- function() {
	params = c(alpha = .55, beta = 0.30, epsilon = 4)
	run.and.display(params, 40, 29, 9)
}


example.95.150.150 <- function() {
	params = c(alpha = .95, beta = 1.50, epsilon = 15)
	run.and.display(params, 21, 23, 1)
}

example.95.1.150 <- function() {
	params = c(alpha = .95, beta = 0.01, epsilon = 13)
	run.and.display(params, 31, 31, 15)
}


example.55.1.50 <- function() {
	params = c(alpha = .55, beta = 0.01, epsilon = 5)
	run.and.display(params, 45, 45, 22)
}



little.example <- function() {
	params = c(alpha = .7, beta = .7, epsilon = 5)
	run.and.display(params, 15, 15, 1)
}



maple.example <- function() {
	params = c(alpha = .68, beta = .38, epsilon = 5)
  run.and.display(params, 21, 19, 4)
}



tiny.bendcost.example <- function() {
	params = c(alpha = .7, beta = 0.01, epsilon = 6.5)
	run.and.display(params, 55, 55, 27)
}



################################################################################
# top level

# These are for invisible edge leading to root. Globals. Not changed.
Stalk.x = 0
Stalk.y = -1


#little.tree = little.example()
#maple.tree = maple.example()
#tree.55.100.50 = example.55.100.50()
#tree.95.150.150 = example.95.150.150()
tree.55.30.50 = example.55.30.50()

#nbt = tiny.bendcost.example()

#example.95.1.150()
#example.55.1.50()
