
# Main Sections:
#     Changing the tree topology: adding nodes, reparenting nodes
#     Tree traversal
#     Filling in costs of veins: weights, bends, accumulation
#     Making grid versions of the tree for display and growth
#     Vein optimisation
#     Adding cells at border
#     Display, as image and text
#     Top level 


################################################################################
# Changing the tree topology: adding nodes, reparenting nodes

make.root <- function() {
	root = list(N=1, pN=-1, cNs=integer(0), x=0, y=0, w=NA, m=NA, f=1, F=1)
	list(root)
}


make.node <- function() {
	list(N=-1, pN=-1, cNs=integer(0), x=NA, y=NA, w=NA, m=NA, f=NA, F=NA)
}


add.child <- function(tree, pN, x, y) {
	node = make.node()
	node$N = length(tree) + 1
	node$pN = pN
	node$x = x
	node$y = y
	tree[[pN]]$cNs = c(tree[[pN]]$cNs, node$N)
	new.tree = c(tree , list(node))
	new.tree
}



# can reparent to existing parent (does nothing)
reparent.node <- function(tree, n, qN) {
	# remove n from current parent's children
	pN = tree[[n]]$pN
	pcNs = tree[[pN]]$cNs
	tree[[pN]]$cNs = pcNs[pcNs != n]
	# make qN the new parent
	tree[[n]]$pN = qN
	tree[[qN]]$cNs = c(tree[[qN]]$cNs, n)
	tree
}


################################################################################
# Tree traversal

collect.subtree.postorder <- function(tree, N, Ns) {
	node = tree[[N]]
	nc = length(node$cNs)
	if (nc > 0) {
		for (ch in 1:nc) {
			Ns = collect.subtree.postorder(tree, node$cNs[ch], Ns)
		}
	}
	Ns = c(Ns, node$N)
	Ns
}


collect.subtree.preorder <- function(tree, N, Ns) {
	node = tree[[N]]
	Ns = c(Ns, node$N)
	nc = length(node$cNs)
	if (nc > 0) {
		for (ch in 1:nc) {
			Ns = collect.subtree.preorder(tree, node$cNs[ch], Ns)
		}
	}
	Ns
}

collect.postorder <- function(tree) {
	Ns = NULL
	Ns = collect.subtree.postorder(tree, 1, Ns)
	Ns
}  


collect.preorder <- function(tree) {
	Ns = NULL
	Ns = collect.subtree.preorder(tree, 1, Ns)
	Ns
}  


################################################################################
# Filling in costs of veins: weights, bends, accumulation

fillin.weights <- function(tree) {
	Ns = collect.postorder(tree)
	for (N in Ns) {
		node = tree[[N]]
		w = 1
		nc = length(node$cNs)
		if (nc > 0) {
			for (ch in 1:nc) {
				cN = node$cNs[ch]
				w = w + tree[[cN]]$w
			}
		}
		tree[[N]]$w = w
	}
	tree
}


bendcost <- function(pdx, pdy, dx, dy) {
	cos.theta = 0
	dotp = pdx*dx + pdy*dy
	if (dotp > 0) {
		dist = sqrt((pdx^2 + pdy^2) * (dx^2 + dy^2))
		cos.theta = dotp / dist
	}
	1 / cos.theta
}





fillin.bends <- function(tree) {
	Ns = collect.preorder(tree)
	for (N in Ns) {
		node = tree[[N]]
		m = 1
		pN = node$pN
		if (pN > 0) {
			x = node$x
			y = node$y
			px = tree[[pN]]$x
			py = tree[[pN]]$y
			# parent coords of root in case no grandparent of N
			ppx = 0
			ppy = -1
			ppN = tree[[pN]]$pN
			if (ppN > 0) {
				ppx = tree[[ppN]]$x
				ppy = tree[[ppN]]$y
			}
			pdx = px - ppx
			pdy = py - ppy
			dx = x - px
			dy = y - py
			b = bendcost(pdx, pdy, dx, dy)
			m = tree[[pN]]$m * b
		}
		tree[[N]]$m = m
	}
	tree
}




fillin.vein.costs <- function(tree, alpha, beta) {
	nnodes = length(tree)
	for (N in 2:nnodes) {
		node = tree[[N]]
		pN = node$pN
		stopifnot(pN > 0)
		dx = node$x - tree[[pN]]$x
		dy = node$y - tree[[pN]]$y
		tree[[N]]$f = node$m^beta * node$w^alpha * sqrt(dx^2 + dy^2)
	}
	Ns = collect.postorder(tree)
	for (N in Ns) {
		node = tree[[N]]
		F = tree[[N]]$f
		nc = length(node$cNs)
		if (nc > 0) {
			for (ch in 1:nc) {
				cN = node$cNs[ch]
				F = F + tree[[cN]]$F
			}
		}
		tree[[N]]$F = F
	}
	tree
}



fillin.all.costs <- function(tree, params) {
	tree = fillin.weights(tree)
	tree = fillin.bends(tree)
	tree = fillin.vein.costs(tree, params["alpha"], params["beta"])
	tree
}


################################################################################
# making grid versions of the tree for both display, and finding
# cells next to border to (possibly) add.

# Note that grid is (row, column) so (y,x) and image has rows increasing
# upwards, unlike matrix
# root is at (0,0) in tree, (1-miny, 1-minx) in grid


tree.bbox <- function(tree) {
	minx = Inf
	maxx = -Inf
	miny = Inf
	maxy = -Inf
	nnodes = length(tree)
	for (n in 1:nnodes) {
		minx = min(minx, tree[[n]]$x)
		maxx = max(maxx, tree[[n]]$x)
		miny = min(miny, tree[[n]]$y)
		maxy = max(maxy, tree[[n]]$y)
	}
	list(minx=minx, maxx=maxx, miny=miny, maxy=maxy, w = 1 + maxx - minx, h = 1 + maxy - miny) 
}

x.y.to.N <- function(grid, bbox, x, y) {
	xx = 1 + x - bbox$minx
	yy = 1 + y - bbox$miny
	n = -1
	if (xx >= 1 & xx <= bbox$w & yy >= 1 & yy <= bbox$h )
	{
		n = grid[yy,xx]
	}
	n
}



tree.as.grid.nodenums <- function(tree, bbox) {
	grid = matrix(-1, nrow=bbox$h, ncol=bbox$w)
	nnodes = length(tree)
	for (n in 1:nnodes) {
		grid[ 1 + tree[[n]]$y - bbox$miny, 1 + tree[[n]]$x - bbox$minx] = tree[[n]]$N
	}
	grid
}


tree.as.grid.Fs <- function(tree, bbox) {
	grid = matrix(0, nrow=bbox$h, ncol=bbox$w)
	nnodes = length(tree)
	maxF = tree[[1]]$F
	for (n in 1:nnodes) {
		grid[ 1 + tree[[n]]$y - bbox$miny, 1 + tree[[n]]$x - bbox$minx] = 0.1 + tree[[n]]$F / maxF
	}
	grid
}


################################################################################
# Vein optimisation


try.reparent.to.x.y <- function(tree, params, grid, bbox, n, x, y) {
	tree2 = tree
	qN = x.y.to.N(grid, bbox, x, y)
	if (qN > 0) {  
		# if qN is in tree, check it is not n, nor an ancestor of n
		loop = FALSE
		nn = qN
		repeat {
			if (nn == n) {
				loop = TRUE
				break
			}
			nn = tree[[nn]]$pN
			if (nn <= 0) {
				# got to root
				break
			}
		}
		if (!loop) {
			tree2 = reparent.node(tree, n, qN)
			tree2 = fillin.all.costs(tree2, params)
		}  
	} 
	tree2
}




tryto.improve.parent <- function(tree, params, grid, bbox, n) {
	node = tree[[n]]
	pN = node$pN
	best.new.tree = tree
	x = tree[[n]]$x
	y = tree[[n]]$y
	pN = tree[[n]]$pN
	
	stopifnot(is.numeric(pN))
	stopifnot(pN>0)
	
	stopifnot(is.numeric(pN))
	stopifnot(pN > 0)
	px = tree[[pN]]$x
	py = tree[[pN]]$y
	minx = min(x, px) - 1
	maxx = max(x, px) + 1
	miny = min(y, py) - 1
	maxy = max(y, py) + 1
	for (xx in minx:maxx) {
		for (yy in miny:maxy) {
			if ((abs(xx-x) <= 1 & abs(yy-y) <= 1)  |  (abs(xx-px) <= 1 & abs(yy-py) <= 1)) {
				tree2 = try.reparent.to.x.y(tree, params, grid, bbox, n, xx, yy)
				if (tree2[[1]]$F < best.new.tree[[1]]$F) {
					best.new.tree = tree2
				}
			}
		}
	}
	best.new.tree
}



tryto.improve.veins <- function(tree, params, grid, bbox) {
	nnodes = length(tree)
	for (n in 2:nnodes) {
		tree = tryto.improve.parent(tree, params, grid, bbox, n)
	}
	tree
}



optimise.veins <- function(tree, params) {
	bbox = tree.bbox(tree)
	grid = tree.as.grid.nodenums(tree, bbox)
	
	repeat {
		F1 = tree[[1]]$F
		tree = tryto.improve.veins(tree, params, grid, bbox)
		tree = fillin.all.costs(tree, params)
		F2 = tree[[1]]$F
		if (F2 == F1) break
	}
	tree
}




################################################################################
# Adding cells at border


best.parent <- function(tree, grid, bbox, params, x, y) {
	pNs = integer(0)
	for (xx in (x-1):(x+1)) {
		for (yy in (y-1):(y+1)) {
			pN = x.y.to.N(grid, bbox, xx, yy)
			if (pN > 0) {
				# (xx,yy) is in tree
				pNs = c(pNs, pN)
			}
		}
	}
	bestp = -1
	bestcost = Inf
	if (length(pNs) > 0) {
		besti = -1
		for (i in 1:length(pNs)) {
			tree2 = add.child(tree, pNs[i], x, y)
			tree2 = fillin.all.costs(tree2, params)
			cost = tree2[[1]]$F
			if (cost < bestcost) {
				bestcost = cost
				besti = i
			}
		}
		if (besti > 0) {
			bestp = pNs[besti]
		}
	}
	list(p=bestp, cost=bestcost)
}



add.cells <- function(tree, params, epsilon) {
	bbox = tree.bbox(tree)
	grid = tree.as.grid.nodenums(tree, bbox)
	new.tree = tree
	minx = bbox$minx - 1
	maxx = bbox$maxx + 1
	miny = bbox$miny - 1  
	maxy = bbox$maxy + 1
	for (x in minx:maxx) {
		for (y in miny:maxy) {
			m = x.y.to.N(grid, bbox, x, y)
			if (m < 0) {
				# (x,y) is not in tree
				best = best.parent(tree, grid, bbox, params, x, y)
				if ((best$p > 0)  &  (best$cost < tree[[1]]$F + epsilon)) {
					new.tree = add.child(new.tree, best$p, x, y)
				}
			}
		}
	}
	new.tree = fillin.all.costs(new.tree, params)
	new.tree
}




################################################################################
# Top level loop: optimise veins, grow leaf, until run out of juice
# 
# repeat {
#   optimise.veins()
#   fillin.all.costs(alpha, beta)
#   new.tree = add.cells(3)
#   if (new.tree no bigger) break
# }


################################################################################
# Display, as image and text

tree.as.image <- function(tree) {
	bbox = tree.bbox(tree)
	grid = tree.as.grid.Fs(tree, bbox)
	x = 1:ncol(grid)
	y = 1:nrow(grid)
	z = t(grid)
	nc = 100
	clrs = rev(hcl.colors(nc, palette = "viridis", alpha = NULL, rev = FALSE, fixup = TRUE))
	clrs = c("#f8f8f8", clrs)
	image(x, y, z, col=clrs)
	nnodes = length(tree)
	xoff = 1 - bbox$minx
	yoff = 1 - bbox$miny
	for (n in 1:nnodes) {
		node = tree[[n]]
		text(node$x+xoff, node$y+yoff, node$N, col="gray")
	}
	for (n in 2:nnodes) {
		node = tree[[n]]
		pnode = tree[[ node$pN ]]
		lines(x=xoff+c(pnode$x, node$x), y=yoff+c(pnode$y, node$y), lwd = sqrt(node$w))
	}
}





tree.nodes.as.matrix <- function(tree) {
	nnodes = length(tree)
	mm = matrix(0, nrow=nnodes, ncol=8)
	colnames(mm) = c("N","pN","x","y","W","m","f","F")
	for (n in 1:nnodes) {
		node = tree[[n]]
		mm[n,] = c(node$N, node$pN, node$x, node$y, node$w, node$m, node$f, node$F)
	} 
	mm
}


print.nodes <- function(tree) {
	mm = tree.nodes.as.matrix(tree)
	print(mm)
}

################################################################################



TheTreeG = make.root()
TheParameters = c(alpha = 1, beta = 1)

epss = seq(from=2, to=10*TheParameters["alpha"], length.out = 40)

for (eps in epss) {
	TheTreeG = add.cells(TheTreeG, TheParameters, eps)
	TheTreeG = fillin.all.costs(TheTreeG, TheParameters)
	TheTreeG = optimise.veins(TheTreeG, TheParameters)
	cat("epsilon = ", eps, "\n")
	if (length(TheTreeG) > 200) {
		break
	}
	tree.as.image(TheTreeG)
}




